#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-08-06
# @Filename: notifier.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import asyncio
import contextlib
import enum


__all__ = ["EventNotifier", "EventListener"]


class EventNotifier(object):
    """A registry of clients to be notified of events.

    Allows to register a listener queue in which to announce events.

    """

    def __init__(self):

        self.listeners = []

    def register_listener(self, listener):
        """Registers a listener.

        Parameters
        ----------
        listener : .EventListener
            An `.EventListener` instance to which to send events for
            processing.

        """

        assert isinstance(listener, EventListener), "invalid listener type."

        if listener not in self.listeners:
            self.listeners.append(listener)

    def remove_listener(self, listener):
        """Removes a listener."""

        if listener not in self.listeners:
            raise ValueError("listener is not registered.")

        self.listeners.remove(listener)

    def notify(self, event, payload={}):
        """Sends an event to all listeners.

        Parameters
        ----------
        event : ~enum.Enum
            An enumeration value belonging to the ``event_class``.
        payload : dict
            A dictionary with the information associated with the event.

        """

        assert isinstance(event, enum.Enum), "event is not an enum."

        for listener in self.listeners:

            if listener.filter_events and event not in listener.filter_events:
                return False

            listener.put_nowait((event, payload))

        return True


class EventListener(asyncio.Queue):
    """An event queue with callbacks.

    Parameters
    ----------
    filter_events : list
        A list of enum values of which to be notified. If `None`,
        all events will be notified.
    autostart : bool
        Whether to start the listener as soon as the object is created.

    """

    def __init__(self, loop=None, filter_events=None, autostart=True):

        asyncio.Queue.__init__(self)

        self.callbacks = []

        self.loop = loop or asyncio.get_running_loop()

        self.filter_events = filter_events
        if self.filter_events and not isinstance(self.filter_events, (list, tuple)):
            self.filter_events = [self.filter_events]

        self.listerner_task = None

        self._event_waiter = None
        self.__events = set()  # A list of events received to be used by wait_for

        if autostart:
            self.listerner_task = self.loop.create_task(self._process_queue())

    async def _process_queue(self):
        """Processes the queue and calls callbacks."""

        while True:
            try:
                event, payload = await self.get()
            except TypeError:
                continue

            for callback in self.callbacks:
                cb = callback(event, payload)
                if asyncio.iscoroutine(cb):
                    self.loop.create_task(cb)

            if self._event_waiter:
                self.__events.add(event)
                self._event_waiter.set()

    async def start_listening(self):
        """Starts the listener task. The queue will be initially purged."""

        if self.listerner_task is not None:
            await self.stop_listening()

        # Purges the queue
        while True:
            try:
                self.get_nowait()
            except asyncio.QueueEmpty:
                break

        self.listerner_task = self.loop.create_task(self._process_queue())

    async def stop_listening(self):
        """Stops the listener task."""

        if self.listerner_task is None:
            return

        self.listerner_task.cancel()

        with contextlib.suppress(asyncio.CancelledError):
            await self.listerner_task

    def register_callback(self, callback):
        """Registers a callback to be called when an event is read.

        Parameters
        ----------
        callback:
            A function or coroutine function to be called. The callback
            receives the event (an enumeration value) as the first argument
            and the payload associated with that event as a dictionary.
            If the callback is a coroutine, it is scheduled as a task.

        """

        if callback not in self.callbacks:
            self.callbacks.append(callback)

    def remove_callback(self, callback):
        """De-registers a callback."""

        if callback in self.callbacks:
            self.callbacks.remove(callback)
        else:
            raise ValueError("callback not registered.")

    async def wait_for(self, events, timeout=None):
        """Blocks until a certain event happens.

        Parameters
        ----------
        events
            The event to wait for. It can be a list of events, in which case
            it returns when any of the events has been seen.
        timeout : float or None
            Timeout in seconds. If `None`, blocks until the event is received.

        Returns
        -------
        result : bool
            Returns a `set` of events received that intersects with the
            ``events`` that were being waited for. Normally this is a single
            event, the first one to be seen, but it can be more than one if
            multiple events that were being waited for happen at the same time.
            Returns `False` if the routine timed out before receiving the event.

        """

        # We need __events to be a list because if two events arrive too close
        # we may miss some of them.
        self._event_waiter = asyncio.Event()
        self.__events = set()

        events = set(events) if isinstance(events, (list, tuple)) else set([events])

        async def _waiter():
            assert self._event_waiter
            while not events.intersection(self.__events):
                await self._event_waiter.wait()
                # Clear the event waiter. If __last_event == event
                # then it doesn't matter. If _last_event != event,
                # this will block in the next loop.
                self._event_waiter.clear()
            return True

        try:
            await asyncio.wait_for(_waiter(), timeout)
            event_inters = events.intersection(self.__events)
            return event_inters
        except asyncio.TimeoutError:
            return False
        finally:
            self._event_waiter.set()
            self._event_waiter = None
            self.__events = set()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: shutter.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import click

from clu.parsers.click import CluCommand

from basecam.exceptions import CameraError

from ..tools import get_cameras


__all__ = ["shutter"]


@click.command(cls=CluCommand)
@click.argument("CAMERAS", nargs=-1, type=str, required=False)
@click.option("--open", "shutter_position", flag_value="open", help="Open the shutter.")
@click.option(
    "--close", "shutter_position", flag_value="close", help="Close the shutter."
)
async def shutter(command, cameras, shutter_position):
    """Controls the camera shutter.

    If called without a shutter position flag, returns the current position
    of the shutter.
    """

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    failed = False
    for camera in cameras:
        if shutter_position is None:
            shutter_now = await camera.get_shutter()
            command.info(
                shutter=dict(
                    camera=camera.name,
                    shutter="open" if shutter_now else "closed",
                )
            )
        else:
            try:
                await camera.set_shutter(shutter_position == "open")
                command.info(
                    shutter=dict(
                        camera=camera.name,
                        shutter="open" if shutter_position else "closed",
                    )
                )
            except CameraError as ee:
                command.error(
                    error=dict(
                        camera=camera.name,
                        error=f"failed commanding shutter: {ee!s}",
                    )
                )
                failed = True

    if failed:
        return command.fail("failed commanding the shutter of one or more cameras.")
    else:
        return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: expose.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from functools import partial

from typing import TYPE_CHECKING

import click

from clu.parsers.click import unique

from basecam.events import CameraEvent
from basecam.exceptions import ExposureError

from ..tools import get_cameras
from .base import camera_parser


if TYPE_CHECKING:
    from basecam import BaseCamera
    from basecam.actor import BasecamCommand

__all__ = ["expose"]


EXPOSURE_STATE = {}


def report_exposure_state(command: BasecamCommand, event, payload):
    global EXPOSURE_STATE

    if event not in CameraEvent:
        return

    name = payload.get("name", "")
    if not name:
        return

    if name not in EXPOSURE_STATE:
        EXPOSURE_STATE[name] = {}

    EXPOSURE_STATE[name].update(payload)

    if event == CameraEvent.EXPOSURE_INTEGRATING:
        state = "integrating"
    elif event == CameraEvent.EXPOSURE_READING:
        state = "reading"
        EXPOSURE_STATE[name].update({"remaining_time": 0.0})
    elif event == CameraEvent.EXPOSURE_DONE:
        state = "done"
        EXPOSURE_STATE[name].update({"remaining_time": 0.0})
    elif event == CameraEvent.EXPOSURE_FAILED:
        state = "failed"
        EXPOSURE_STATE[name].update(
            {
                "remaining_time": 0.0,
                "current_stack": 0,
                "n_stack": 0,
            }
        )
    elif event == CameraEvent.EXPOSURE_IDLE:
        state = "idle"
        EXPOSURE_STATE[name].update(
            {
                "remaining_time": 0.0,
                "current_stack": 0,
                "n_stack": 0,
                "exptime": 0,
                "image_type": "NA",
            }
        )
    elif event == CameraEvent.EXPOSURE_WRITTEN:
        command.info(
            filename={
                "camera": name,
                "filename": payload.get("filename", "UNKNOWN"),
            }
        )
        return
    elif event == CameraEvent.EXPOSURE_POST_PROCESSING:
        command.info(
            state="post_processing",
            image_type=EXPOSURE_STATE[name].get("image_type", "NA"),
        )
        return
    elif event == CameraEvent.EXPOSURE_POST_PROCESS_FAILED:
        command.warning(
            state="post_process_failed",
            image_type=EXPOSURE_STATE[name].get("image_type", "NA"),
            error=payload.get("error", "Error unknown"),
        )
        return
    else:
        return

    command.info(
        exposure_state={
            "camera": name,
            "state": state,
            "image_type": EXPOSURE_STATE[name].get("image_type", "NA"),
            "remaining_time": EXPOSURE_STATE[name].get("remaining_time", 0),
            "exposure_time": EXPOSURE_STATE[name].get("exptime", 0),
            "current_stack": EXPOSURE_STATE[name].get("current_stack", 0),
            "n_stack": EXPOSURE_STATE[name].get("n_stack", 0),
        }
    )


async def expose_one_camera(
    command,
    camera,
    exptime,
    image_type,
    stack,
    filename,
    num,
    no_postprocess,
    **extra_kwargs,
):
    command.info(text=f"exposing camera {camera.name!r}")
    obj = click.get_current_context().obj
    try:
        exposure = await camera.expose(
            exptime,
            image_type=image_type,
            stack=stack,
            filename=filename,
            num=num,
            postprocess=not no_postprocess,
            write=True,
            **extra_kwargs,
        )
        post_process_cb = obj.get("post_process_callback", None)
        if post_process_cb:
            await post_process_cb(command, exposure)
        return True
    except ExposureError as ee:
        command.error(error={"camera": camera.name, "error": str(ee)})
        return False


@camera_parser.command()
@click.argument("CAMERAS", nargs=-1, type=str, required=False)
@click.argument("EXPTIME", type=float, required=False)
@click.option(
    "--object",
    "image_type",
    flag_value="object",
    default=True,
    show_default=True,
    help="Takes an object exposure.",
)
@click.option(
    "--flat",
    "image_type",
    flag_value="flat",
    show_default=True,
    help="Takes a flat exposure.",
)
@click.option(
    "--dark",
    "image_type",
    flag_value="dark",
    show_default=True,
    help="Takes a dark exposure.",
)
@click.option(
    "--bias",
    "image_type",
    flag_value="bias",
    show_default=True,
    help="Takes a bias exposure.",
)
@click.option(
    "--filename",
    "-f",
    type=str,
    default=None,
    show_default=True,
    help="Filename of the imaga to save.",
)
@click.option(
    "-n",
    "--num",
    type=int,
    default=None,
    help="Sequence number for this exposure filename.",
)
@click.option(
    "--stack",
    "-s",
    type=int,
    default=1,
    show_default=True,
    help="Number of images to stack.",
)
@click.option(
    "-c",
    "--count",
    type=int,
    default=1,
    help="Number of exposures to take.",
)
@click.option(
    "--no-postprocess",
    is_flag=True,
    help="Skip the post-process step, if defined.",
)
@unique()
async def expose(
    command: BasecamCommand,
    cameras: list[BaseCamera],
    exptime: float | None = None,
    image_type: str = "object",
    filename: str | None = None,
    num: int | None = None,
    stack: int | None = None,
    count: int = 1,
    no_postprocess: bool = False,
    **exposure_kwargs,
):
    """Exposes and writes an image to disk."""

    for _ in range(count):
        cameras = get_cameras(command, cameras=cameras, fail_command=True)
        if not cameras:  # pragma: no cover
            return

        if image_type == "bias":
            if exptime and exptime > 0:
                command.warning("Setting exposure time for bias to 0 seconds.")
            exptime = 0.0

        if exptime is None:
            return command.fail("Exposure time not provided.")

        if filename and len(cameras) > 1:
            return command.fail("--filename can only be used with a single camera.")

        report_exposure_state_partial = partial(report_exposure_state, command)

        command.actor.listener.register_callback(report_exposure_state_partial)
        jobs = []
        for camera in cameras:
            jobs.append(
                asyncio.create_task(
                    expose_one_camera(
                        command,
                        camera,
                        exptime,
                        image_type,
                        stack,
                        filename,
                        num,
                        no_postprocess,
                        **exposure_kwargs,
                    )
                )
            )

        try:
            results = await asyncio.gather(*jobs)
        except Exception as err:
            command.error(err)
            results = (False,)
        finally:
            # Wait a bit to allow leftover messages sent by the notifier to be output.
            # TODO: this can be improved by checking EXPOSURE_STATE and confirming that
            # all EXPOSURE_WRITTEN has been output for all the cameras (assuming no
            # failures).
            await asyncio.sleep(0.5)

            # Remove the listener.
            command.actor.listener.remove_callback(report_exposure_state_partial)

        if not all(results):
            return command.fail("One or more cameras failed to expose.")
        else:
            for camera in cameras:
                # Reset cameras to idle
                report_exposure_state(
                    command,
                    CameraEvent.EXPOSURE_IDLE,
                    {"name": camera.name},
                )

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-10-04
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# flake8: noqa


from . import commands
from .actor import *
from .tools import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: set_default.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import click

from .base import camera_parser


__all__ = ["set_default"]


@camera_parser.command()
@click.argument("CAMERAS", nargs=-1, type=str)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Forces a camera to be set as default even if it is not connected.",
)
async def set_default(command, cameras, force):
    """Set default cameras."""

    camera_system = command.actor.camera_system

    camera_names = []
    for camera in cameras:
        for camera_split in camera.split(","):
            if camera_split not in camera_names:
                camera_names.append(camera_split)

    for camera_name in camera_names:
        camera = camera_system.get_camera(name=camera_name)

        if not camera:
            msg = f"camera {camera_name} is not connected."
            if force:
                command.write("w", text=msg)
            else:
                return command.fail(
                    camera=camera.name, text=f"camera {camera_name} is not connected."
                )

    command.actor.set_default_cameras(camera_names)

    command.finish(default_cameras=command.actor.default_cameras)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-10
# @Filename: fits.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import functools
from copy import copy

from typing import Any, Dict, List, Optional, TypeVar, Union

import astropy.io.fits
import astropy.table
import numpy

import basecam.exposure

from ..exceptions import CardError
from .card import DEFAULT_CARDS, Card, CardGroup, MacroCard


try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


__all__ = [
    "FITSModel",
    "Extension",
    "HeaderModel",
]


Exposure = basecam.exposure.Exposure

ImageHDUType = Union[
    astropy.io.fits.ImageHDU,
    astropy.io.fits.CompImageHDU,
    astropy.io.fits.PrimaryHDU,
]
_CardTypes = Union[Card, CardGroup, MacroCard, str, tuple, list, None]
T = TypeVar("T", bound="FITSModel")


class FITSModel(list):
    """A model representing a FITS image.

    This model defines the data and header for each extension in a FITS
    image. The model can be evaluated for given `.Exposure`, returning a
    fully formed astropy `~astropy.io.fits.HDUList`.

    Parameters
    ----------
    extension
        A list of HDU extensions defined as `.Extension` objects. If none is
        defined, a single, basic extension is added.
    context
        A default context to pass to the header model when evaluating it.
    """

    def __init__(
        self,
        extensions: Optional[List[Extension]] = None,
        context: Dict[str, Any] = {},
    ):

        self.context: Dict[str, Any] = context

        extensions = extensions or []
        list.__init__(self, extensions)

        if len(self) == 0:
            self.append(Extension(data="raw", name="DATA"))

    def copy(self: T) -> T:
        return copy(self)

    def to_hdu(
        self,
        exposure: Exposure,
        context: Dict[str, Any] = {},
    ) -> astropy.io.fits.HDUList:
        """Returns an `~astropy.io.fits.HDUList` from an exposure.

        Parameters
        ----------
        exposure
            The exposure for which the FITS model is evaluated.
        context
            A dictionary of parameters used to fill the card replacement
            fields. Updates the default context.

        Returns
        -------
        hdulist
            A list of HDUs, each one define by its corresponding `.Extension`
            instance. The first extension is created as a
            `~astropy.io.fits.PrimaryHDU` unless it is compressed, in which
            case an empty primary HDU is prepended.
        """

        hdus = []

        ucontext = self.context.copy()
        ucontext.update(context)

        for i, ext in enumerate(self):
            if i == 0:
                primary = True
                if ext.compressed:
                    # Compressed HDU cannot be primary.
                    hdus.append(astropy.io.fits.PrimaryHDU())
                    primary = False
            else:
                primary = False
            hdus.append(ext.to_hdu(exposure, primary=primary, context=ucontext))

        return astropy.io.fits.HDUList(hdus)


class Extension(object):
    """A model for a FITS extension.

    Parameters
    ----------
    data
        The data for this FITS extension. Can be an array or a macro string
        indicating the type of data to store in the extension. Available macros
        are: ``'raw'`` or `None` for the raw image, or ``'none'`` for empty data.
    header_model
        A `.HeaderModel` for this extension.
    name
        The name of the HDU.
    compressed
        If `False`, the extension data will not be compressed. Otherwise, a
        string with one of the ``compression_type`` in
        `~astropy.io.fits.CompImageHDU`.
    compression_params
        Additional parameters to be passed to `~astropy.io.fits.CompImageHDU` if the
        extension is compressed.
    """

    __VALID_DATA_VALUES = ["raw", "none", True, False]

    def __init__(
        self,
        data: Union[Literal["raw"], Literal["none"], None, bool, numpy.ndarray] = None,
        header_model: Optional[HeaderModel] = None,
        name: Optional[str] = None,
        compressed: Union[bool, str] = False,
        compression_params: dict[str, Any] = {},
    ):

        if isinstance(data, numpy.ndarray):
            self.data = data
        else:
            assert data is None or data in self.__VALID_DATA_VALUES, "invalid data"
            self.data = data

        self.header_model = header_model

        self.name = name

        self.compressed = compressed
        if self.compressed is True:
            self.compressed = "GZIP_2"
        self._compression_params = compression_params

    def __repr__(self):
        return f"<Extension (name={self.name!r}, compressed={self.compressed})>"

    def to_hdu(
        self,
        exposure: Exposure,
        primary: bool = False,
        context: Dict[str, Any] = {},
    ) -> ImageHDUType:
        """Evaluates the extension as an HDU.

        Parameters
        ----------
        exposure
            The exposure for which we want to evaluate the extension.
        primary
            Whether this is the primary extension.
        context
            A dictionary of arguments used to evaluate the parameters in
            the extension.

        Returns
        -------
        hdu
            An `~astropy.io.fits.ImageHDU` with the data and header evaluated
            for ``exposure``, or `~astropy.io.fits.CompImageHDU` if
            ``compressed=True``.
        """

        if self.compressed:
            HDUClass = astropy.io.fits.CompImageHDU
            if isinstance(self.compressed, str):
                HDUClass = functools.partial(
                    HDUClass,
                    compression_type=self.compressed,
                    **self._compression_params,
                )
        else:
            if primary:
                HDUClass = astropy.io.fits.PrimaryHDU
            else:
                HDUClass = astropy.io.fits.ImageHDU

        if not primary:
            HDUClass = functools.partial(HDUClass, name=self.name)

        data = self.get_data(exposure, primary=primary)

        # Create the HDU without a header first to allow astropy to create a
        # basic header (for example, if HDUClass is CompImageHDU this will add
        # the BITPIX keyword). Then append our header.
        hdu = HDUClass(data=data, header=None)

        if self.header_model:
            hdu.header.extend(self.header_model.to_header(exposure, context=context))

        return hdu

    def get_data(
        self,
        exposure: Exposure,
        primary: bool = False,
    ) -> Union[numpy.ndarray, None]:
        """Returns the data as a numpy array."""

        if isinstance(self.data, str):
            if self.data == "raw":
                data = exposure.data
            else:
                data = None
        elif isinstance(self.data, numpy.ndarray):
            data = self.data
        elif self.data is None or self.data is True:
            data = exposure.data
        elif self.data is False:
            data = None
        else:
            raise ValueError(f"Invalid data value {self.data!r}")

        return data


class HeaderModel(list):
    """A model defining the header of an HDU.

    Parameters
    ----------
    cards
        A list of `.Card`, `.CardGroup`, or `MacroCard` instances. It can also be a
        string with the name of a :ref:`default card <default-cards>`.

    Examples
    --------
    >>> header_model = HeaderModel([Card('TELESCOP', 'APO-2.5', 'The telescope'),
                                    Card('OBSERVATORY', 'APO'),
                                    'EXPTIME',
                                    Card('camname', '{(camera).name}')])
    """

    def __init__(self, cards: List[_CardTypes] = []):

        cards = [self._process_input(card) for card in cards]
        list.__init__(self, cards)

    def __repr__(self):

        return f"<HeaderModel {list.__repr__(self)!s}>"

    def _process_input(self, input_card: _CardTypes) -> _CardTypes:
        """Processes the input and converts it into a valid card."""

        if isinstance(input_card, (Card, CardGroup, MacroCard)):
            return input_card
        elif isinstance(input_card, str):
            if input_card not in DEFAULT_CARDS:
                raise CardError(f"{input_card} is not a default card.")
            return Card(input_card.upper())
        elif isinstance(input_card, (tuple, list)):
            name, value, *other = input_card
            return Card(name, value=value, comment=other[0] if other else "")
        elif input_card is None:
            return None
        else:
            raise CardError(f"invalid input {input_card!r}")

    def append(self, card: _CardTypes):
        list.append(self, self._process_input(card))

    def insert(self, idx: int, card: _CardTypes):
        list.insert(self, idx, self._process_input(card))

    def to_header(
        self,
        exposure: Exposure,
        context: Dict[str, Any] = {},
    ) -> astropy.io.fits.Header:
        """Evaluates the header model for an exposure and returns a header.

        Parameters
        ----------
        exposure
            The exposure for which we want to evaluate the model.
        context
            A dictionary of arguments used to evaluate the parameters in
            the model.

        Returns
        -------
        header
            A `~astropy.io.fits.Header`, created by evaluating the model for
            the input exposure.
        """

        header = astropy.io.fits.Header()

        for card in self:

            processed_card = self._process_input(card)
            if processed_card is not None:
                if isinstance(processed_card, Card):
                    header.append(processed_card.evaluate(exposure, context=context))
                elif isinstance(processed_card, (CardGroup, MacroCard)):
                    header += processed_card.to_header(exposure, context=context)

        return header

    def describe(self):
        """Returns a table-like representation of the header model."""

        rows = []
        for card in self:
            if isinstance(card, Card):
                rows.append((card.name.upper(), card.value, card.comment))
            elif isinstance(card, CardGroup):
                for card_ in card:
                    rows.append((card_.name.upper(), card_.value, card_.comment))
            elif isinstance(card, MacroCard):
                rows.append(("### MACRO", card.__class__.__name__, ""))
            else:
                raise CardError(f"Invalid card {card}. This should not have happened.")

        return astropy.table.Table(rows=rows, names=["name", "value", "comment"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: binning.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import click

from clu.parsers.click import CluCommand

from basecam.exceptions import CameraError

from ..tools import get_cameras


__all__ = ["binning"]


@click.command(cls=CluCommand)
@click.argument("CAMERAS", nargs=-1, type=str, required=False)
@click.argument("BINNING", nargs=2, type=int, required=False)
async def binning(command, cameras, binning):
    """Controls the camera binning.

    If called without a binning value, returns the current value.
    """

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    failed = False
    for camera in cameras:

        if not binning:
            binning = list(await camera.get_binning())
            command.info(
                binning=dict(
                    camera=camera.name,
                    horizontal=binning[0],
                    vertical=binning[1],
                )
            )
        else:
            try:
                await camera.set_binning(*binning)
                command.info(
                    binning=dict(
                        camera=camera.name,
                        horizontal=binning[0],
                        vertical=binning[1],
                    )
                )
            except (CameraError, AssertionError) as ee:
                command.error(error=dict(camera=camera.name, error=str(ee)))
                failed = True

    if failed:
        return command.fail("failed to set binning for one or more cameras.")
    else:
        return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-09
# @Filename: mixins.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import abc
import asyncio

from typing import Callable, Union

from basecam.camera import BaseCamera

from .events import CameraEvent
from .utils import cancel_task


__all__ = ["ShutterMixIn", "ExposureTypeMixIn", "CoolerMixIn", "ImageAreaMixIn"]


class ShutterMixIn(object, metaclass=abc.ABCMeta):
    """A mixin that provides manual control over the shutter."""

    @abc.abstractmethod
    async def _set_shutter_internal(self, shutter_open):
        """Internal method to set the position of the shutter."""

        raise NotImplementedError

    @abc.abstractmethod
    async def _get_shutter_internal(self):
        """Internal method to get the position of the shutter."""

        raise NotImplementedError

    async def set_shutter(self, shutter, force=False):
        """Sets the position of the shutter.

        Parameters
        ----------
        shutter : bool
            If `True` moves the shutter open, otherwise closes it.
        force : bool
            Normally, a call is made to `.get_shutter` to determine if the
            shutter is already in the commanded position. If it is, the
            shutter is not commanded to move. ``force=True`` sends the
            move command regardless of the internal status of the shutter.

        """

        current_status = await self.get_shutter()
        if current_status == shutter and not force:
            return

        return await self._set_shutter_internal(shutter)

    async def open_shutter(self):
        """Opens the shutter (alias for ``set_shutter(True)``)."""

        return await self.set_shutter(True)

    async def close_shutter(self):
        """Opens the shutter (alias for ``set_shutter(False)``)."""

        return await self.set_shutter(False)

    async def get_shutter(self):
        """Gets the position of the shutter."""

        return await self._get_shutter_internal()


class ExposureTypeMixIn(object):
    """Methods to take exposures with different image_types."""

    expose: Callable

    async def bias(self, **kwargs):
        """Take a bias image."""

        kwargs.pop("image_type", None)

        return await self.expose(0.0, image_type="bias", **kwargs)

    async def dark(self, exp_time: float, **kwargs):
        """Take a dark image."""

        kwargs.pop("image_type", None)

        return await self.expose(exp_time, image_type="dark", **kwargs)

    async def flat(self, exp_time: float, **kwargs):
        """Take a flat image."""

        kwargs.pop("image_type", None)

        return await self.expose(exp_time, image_type="flat", **kwargs)

    async def object(
        self: Union[BaseCamera, ExposureTypeMixIn],
        exp_time: float,
        **kwargs,
    ):
        """Take a science image."""

        kwargs.pop("image_type", None)

        return await self.expose(exp_time, image_type="object", **kwargs)


class CoolerMixIn(object, metaclass=abc.ABCMeta):
    """Methods to control the cooling system of the camera."""

    _set_temperature_task = None

    notify: Callable

    async def set_temperature(self, temperature):
        """Sets a new temperature goal for the camera.

        Emits a `~.CameraEvent.NEW_SET_POINT` event when the new temperature
        is set. The coroutine blocks until the temperature has been reached
        (at which point it emits `~.CameraEvent.SET_POINT_REACHED`) or until
        the set point changes.

        Parameters
        ----------
        temperature : float
            The goal temperature, in degrees Celsius.

        """

        async def _wait_for_temp(temp):
            while abs(await self.get_temperature() - temp) > 0.1:
                await asyncio.sleep(0.5)
            self.notify(CameraEvent.SET_POINT_REACHED, {"temperature": temp})

        # If there is already a set point, cancel it (this does not cancel
        # the cooler changing the temperature).
        await cancel_task(self._set_temperature_task)

        await self._set_temperature_internal(temperature)

        self.notify(CameraEvent.NEW_SET_POINT, {"temperature": temperature})

        loop = asyncio.get_event_loop()
        self._set_temperature_task = loop.create_task(_wait_for_temp(temperature))
        await self._set_temperature_task

    @abc.abstractmethod
    async def _set_temperature_internal(self, temperature):
        """Internal method to set the camera temperature.

        This method should return immediately after setting the new temperature
        or raise `.CameraError` if there is a problem.

        """

        raise NotImplementedError

    async def get_temperature(self):
        """Returns the temperature of the camera."""

        return await self._get_temperature_internal()

    @abc.abstractmethod
    async def _get_temperature_internal(self):
        """Internal method to get the camera temperature.

        If the camera can report multiple temperatures, this method must
        return the temperature that the cooler modifies. Other temperature
        can be reported in the status. Must raise `.CameraError` if there is
        a problem.

        """

        raise NotImplementedError


class ImageAreaMixIn(object, metaclass=abc.ABCMeta):
    """Allows to select the image area and binning."""

    async def get_image_area(self):
        """Returns the imaging area as 1-indexed ``(x0, x1, y0, y1)``."""

        return await self._get_image_area_internal()

    @abc.abstractmethod
    async def _get_image_area_internal(self):
        """Internal method to return the image area."""

        raise NotImplementedError

    async def set_image_area(self, area=None):
        """Sets the image area.

        Parameters
        ----------
        area : tuple
            The image area to set as 1-indexed ``(x0, x1, y0, y1)``.
            If not provided, restores the full image area.

        """

        return await self._set_image_area_internal(area=area)

    @abc.abstractmethod
    async def _set_image_area_internal(self, area=None):
        """Internal method to set the image area.

        If ``area=None`` must restore the full image area. In case of error,
        must raise `.CameraError`.

        """

        raise NotImplementedError

    async def get_binning(self):
        """Returns the horizontal and vertical binning as ``(hbin, vbin)``."""

        return await self._get_binning_internal()

    @abc.abstractmethod
    async def _get_binning_internal(self):
        """Internal method to return the binning."""

        raise NotImplementedError

    async def set_binning(self, hbin=1, vbin=None):
        """Sets the binning.

        Parameters
        ----------
        hbin : int
            Horizontal binning.
        vbin : int
            Vertical binning. If not provided, same as ``hbin``.

        """

        assert isinstance(hbin, int), "hbin must be an integer."

        vbin = vbin or hbin
        assert isinstance(vbin, int), "vbin must be an integer."

        return await self._set_binning_internal(hbin, vbin)

    @abc.abstractmethod
    async def _set_binning_internal(self, hbin, vbin):
        """Internal method to set the binning.

        In case of error it must raise `.CameraError`.

        """

        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2019-10-03
# @Filename: events.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import enum


__all__ = ["CameraSystemEvent", "CameraEvent"]


class CameraSystemEvent(enum.Enum):
    """Enumeration of camera system events."""

    CAMERA_ADDED = enum.auto()
    CAMERA_REMOVED = enum.auto()


class CameraEvent(enum.Enum):
    """Enumeration of camera events."""

    CAMERA_CONNECTED = "connected"
    CAMERA_CONNECT_FAILED = "connect_failed"
    CAMERA_DISCONNECTED = "disconnected"
    CAMERA_DISCONNECT_FAILED = "disconnect_failed"

    EXPOSURE_IDLE = "idle"
    EXPOSURE_FLUSHING = "flushing"
    EXPOSURE_INTEGRATING = "integrating"
    EXPOSURE_READING = "reading"
    EXPOSURE_READ = "read"
    EXPOSURE_DONE = "done"
    EXPOSURE_FAILED = "failed"
    EXPOSURE_WRITING = "writing"
    EXPOSURE_WRITTEN = "written"
    EXPOSURE_POST_PROCESSING = "post_processing"
    EXPOSURE_POST_PROCESS_DONE = "post_process_done"
    EXPOSURE_POST_PROCESS_FAILED = "post_process_failed"

    NEW_SET_POINT = "new_set_point"
    SET_POINT_REACHED = "set_point_reached"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: area.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import click

from clu.legacy import LegacyActor
from clu.parsers.click import CluCommand

from basecam.exceptions import CameraError

from ..tools import get_cameras


__all__ = ["area"]


def report_area(command, camera, area):
    """Reports the area, taking into account the type of actor."""

    actor = command.actor

    # Force area to be a list so that JSONSchema array accepts it. This will be
    # fixed in CLU 0.6.3.
    area = list(area)

    if isinstance(actor, LegacyActor):
        command.info(area=[camera.name, *area])
    else:
        command.info(message={"area": dict(camera=camera.name, area=area)})


@click.command(cls=CluCommand)
@click.argument("CAMERAS", nargs=-1, type=str, required=False)
@click.argument("AREA", nargs=4, type=int, required=False)
@click.option(
    "--reset",
    "-r",
    is_flag=True,
    default=False,
    show_default=True,
    help="Restores the original image area.",
)
async def area(command, cameras, area, reset):
    """Controls the camera image area.

    If called without an image area value, returns the current value.
    The image area must have the format (x0, x1, y0, y1) and be 1-indexed.
    """

    cameras = get_cameras(command, cameras=cameras, fail_command=True)
    if not cameras:  # pragma: no cover
        return

    failed = False
    for camera in cameras:

        if not area and reset is False:
            report_area(command, camera, tuple(await camera.get_image_area()))
        else:
            if reset:
                area = None
            try:
                await camera.set_image_area(area)
                report_area(command, camera, tuple(await camera.get_image_area()))
            except CameraError as ee:
                command.error(error=dict(camera=camera.name, error=str(ee)))
                failed = True

    if failed:
        return command.fail("failed to set binning for one or more cameras.")
    else:
        return command.finish()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['basecam', 'basecam.actor', 'basecam.actor.commands', 'basecam.models']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=4.0,<5.0',
 'numpy>=1.17,<2.0',
 'sdss-clu>=1.6.0,<2.0.0',
 'sdsstools>=0.4.0']

extras_require = \
{':python_version <= "3.7"': ['typing-extensions>=3.7.4,<4.0.0']}

setup_kwargs = {
    'name': 'sdss-basecam',
    'version': '0.6.1',
    'description': 'A base library for camera wrapping and actor.',
    'long_description': 'basecam\n=======\n\n![Versions](https://img.shields.io/badge/python-3.8-blue)\n[![PyPI version](https://badge.fury.io/py/sdss-basecam.svg)](https://badge.fury.io/py/sdss-basecam)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Documentation Status](https://readthedocs.org/projects/sdss-basecam/badge/?version=latest)](https://sdss-basecam.readthedocs.io/en/latest/?badge=latest)\n[![Build](https://img.shields.io/github/workflow/status/sdss/basecam/Test)](https://github.com/sdss/basecam/actions)\n[![Coverage Status](https://codecov.io/gh/sdss/basecam/branch/master/graph/badge.svg)](https://codecov.io/gh/sdss/basecam)\n\n``basecam`` provides a wrapper around CCD camera APIs with an SDSS-style TCP/IP actor. The main benefits of using `basecam` are:\n\n- Simplifies the creation of production-level camera libraries by providing all the common boilerplate so that you only need to focus on implementing the parts that are specific to your camera API.\n- Provides a common API regardless of the underlying camera being handled.\n- Powerful event handling and notification.\n',
    'author': 'José Sánchez-Gallego',
    'author_email': 'gallegoj@uw.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sdss/basecam',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-02-14
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from .area import area
from .base import camera_parser, list_
from .binning import binning
from .expose import expose
from .reconnect import reconnect
from .set_default import set_default
from .shutter import shutter
from .status import status
from .temperature import temperature


_MIXIN_TO_COMMANDS = {
    "ShutterMixIn": [shutter],
    "CoolerMixIn": [temperature],
    "ImageAreaMixIn": [binning, area],
}
